"""IsIdentifiable module."""
from uuid import UUID, uuid4
from pysetl.utils import pretty


class IsIdentifiable:
    """Mixin class to provide an UUID to child object."""

    def __init__(self) -> None:
        self.uuid: UUID = uuid4()

    @classmethod
    def name(cls) -> str:
        """Return the name of class in a pretty format."""
        return pretty(cls)



"""
CanPartition module.

Connectors that inherit CanPartition should be able to partition the output by
the given columns on the file system
"""
from abc import ABC, abstractmethod
from pyspark.sql import DataFrame
from typing_extensions import Self


class CanPartition(ABC):
    """CanPartition mixin class."""

    @abstractmethod
    def list_partitions(self: Self) -> list[str]:
        """List all available partitions."""

    @abstractmethod
    def read_partitions(self, partitions: list[str]) -> DataFrame:
        """Read partitions from data source."""



"""CsvConfig module."""
from typing_extensions import Self
from pysetl.enums import FileStorage
from .file_config import FileConfig, FileConfigModel


class CsvConfigModel(FileConfigModel):
    """CsvConfig validator."""

    inferSchema: str = "true"
    header: str = "true"
    delimiter: str = ","


class CsvConfig(FileConfig):
    """CsvConfig class."""

    @property
    def config(self: Self) -> CsvConfigModel:
        """Return validated dataset."""
        _config: dict = self.params | {"storage": FileStorage.CSV}

        return CsvConfigModel(**_config)



"""
Provides the FileConnector for handling operations on various file-based data
sources within the pysetl.connector package.
"""
from urllib.parse import urlparse, ParseResult
from pathlib import Path
from typing_extensions import Self
from pyarrow.fs import (  # type: ignore
    S3FileSystem, FileSystem,  # type: ignore
    LocalFileSystem, FileSelector  # type: ignore
)
from pyspark.sql import DataFrame, DataFrameReader, DataFrameWriter
from pyspark.sql.types import StructType
from pysetl.config.file_config import FileConfig, FileConfigModel
from pysetl.utils.mixins import HasReader, HasWriter, CanDrop, CanPartition
from pysetl.utils.exceptions import InvalidConfigException
from .connector import Connector


class FileConnector(Connector, HasReader, HasWriter, CanDrop, CanPartition):
    """
    Connector to data files.

    This connector can handle file input/outputs on the file system.
    """

    def __init__(self: Self, options: FileConfig) -> None:
        """
        Initiallize for FileConnector.

        A file connector requieres a FileConfig
        """
        self.config: FileConfigModel = options.config
        self.reader_config = options.reader_config
        self.writer_config = options.writer_config

    @property
    def _reader(self: Self) -> DataFrameReader:
        return self.spark.read.options(**self.reader_config)

    def __check_partitions(self: Self, data: DataFrame) -> bool:
        return all(x in data.columns for x in self.config.partition_by) \
            if self.config.partition_by \
            else True

    def _writer(self: Self, data: DataFrame) -> DataFrameWriter:
        if not self.__check_partitions(data):
            raise InvalidConfigException(
                "Partition columns in configuration not in data"
            )

        if isinstance(self.config.data_schema, StructType):
            ordered_data = data.select([
                x.name
                for x
                in self.config.data_schema
            ])
        else:
            ordered_data = data

        return (
            ordered_data.write
            .mode(self.config.savemode.value)
            .options(**self.writer_config)
            .partitionBy(*self.config.partition_by)
            .format(self.config.storage.value)
        )

    @property
    def uri(self: Self) -> ParseResult:
        """Parse string path into a URI."""
        return urlparse(self.config.path)

    @property
    def filesystem(self) -> FileSystem:
        """Get current FS based on URI and configuration."""
        if self.uri.scheme == "s3":
            if not self.config.aws_credentials:
                raise InvalidConfigException("No S3 credentials provided")

            return S3FileSystem(
                access_key=self.config.aws_credentials.access_key,
                secret_key=self.config.aws_credentials.secret_key,
                endpoint_override=f"{self.uri.hostname}:{self.uri.port}",
                allow_bucket_creation=True,
                allow_bucket_deletion=True
            )

        fs, _ = FileSystem.from_uri(uri=self.uri.geturl())  # type: ignore

        return fs

    @property
    def absolute_path(self) -> str:
        """Returns absolute path."""
        return str(Path(self.uri.path).absolute()) \
            if isinstance(self.filesystem, LocalFileSystem) \
            else self.uri.path

    @property
    def base_path(self) -> str:
        """Returns base path."""
        return str(Path(self.absolute_path).parent)

    @property
    def has_wildcard(self: Self) -> bool:
        """Verifies if path has a wildcard."""
        return Path(self.absolute_path).name.endswith("*")

    def list_partitions(self: Self) -> list[str]:
        """Return all found partitions."""
        selector = FileSelector(self.absolute_path)

        return [
            x.path
            for x
            in self.filesystem.get_file_info(selector)
            if x.type.name == "Directory"
        ]

    def write(self: Self, data: DataFrame) -> None:
        """Write data into FileSystem."""
        if self.has_wildcard:
            raise InvalidConfigException("Can't write to wildcard path")

        self._writer(data).save(self.absolute_path)

    def drop(self: Self) -> None:
        """Remove data files from the table directory."""
        self.filesystem.delete_dir(self.absolute_path)

    def read(self: Self) -> DataFrame:
        """Read data from FileSystem into DataFrame."""
        return (
            self._reader
            .format(self.config.storage.value)
            .load(self.absolute_path)
        )

    def read_partitions(self, partitions: list[str]) -> DataFrame:
        """Read specific partitions from data source."""
        return (
            self._reader
            .option("basePath", self.base_path)
            .format(self.config.storage.value)
            .load(partitions)
        )



"""CanUpdateMixin module."""
from abc import abstractmethod, ABC
from pyspark.sql import DataFrame


class CanUpdate(ABC):
    """Abstract mixin to add update functionality to a connector."""

    @abstractmethod
    def update(self, dataframe: DataFrame, columns) -> None:
        """
        Update storage data.

        All matching data given some columns will be updated. Otherwise, will
        be created
        """



"""get_signature module."""
from inspect import signature


def get_signature(function) -> list[str]:
    """Get the expected parameters from a function."""
    return list((
        x
        for x, y
        in signature(function).parameters.items()
    ))



"""
Defines the abstract base class for repositories in the
pysetl.storage.repository package.
"""
from abc import ABC, abstractmethod
from typing import Generic, TypeVar
from typedspark import DataSet, Schema
from typing_extensions import Self


T = TypeVar("T", bound=Schema)


class BaseRepository(Generic[T], ABC):
    """A Repository is an abstract high-level type-safe layer over a
    BaseConnector."""
    @abstractmethod
    def load(self: Self) -> DataSet[T]:
        """Read all data from data storage"""

    @abstractmethod
    def save(self: Self, data: DataSet[T]) -> Self:
        """Save data."""



"""JsonConfig module."""
from typing_extensions import Self
from pysetl.enums import FileStorage
from .file_config import FileConfig, FileConfigModel


class JsonConfigModel(FileConfigModel):
    """JsonConfig validator."""


class JsonConfig(FileConfig):
    """JsonConfig class."""

    @property
    def config(self: Self) -> JsonConfigModel:
        """Returns validated JSON configuration."""
        _config: dict = self.params | {"storage": FileStorage.JSON}

        return JsonConfigModel(**_config)



"""Factory class definition."""
from __future__ import annotations
from abc import abstractmethod, ABC
from inspect import getmembers
from typing import TypeVar, Generic, Optional, get_args, get_origin
from typing_extensions import Self, get_original_bases
from pysetl.utils.mixins import (
    HasLogger, IsIdentifiable, IsWritable
)
from pysetl.workflow.deliverable import Deliverable
from pysetl.workflow.delivery import Delivery
from pysetl.workflow.delivery_type import DeliveryType
from pysetl.utils.pretty import pretty


T = TypeVar("T")


class Factory(
    IsIdentifiable,
    Generic[T],
    HasLogger,
    IsWritable,
    ABC
):
    """
    A Factory is a structured object to ETL data.

    A Factory could be used to manipulate data.
    A Factory is able to read data from a data source, process/transform it
    and write it back to the storage
    """

    consumers: set[type[Factory]] = set()
    delivery_id: str = ""

    def __post_init__(self: Self) -> None:
        """Initialize UUID from IsIdentifiable."""
        IsIdentifiable.__init__(self)  # pragma: no cover

    def __str__(self: Self) -> str:
        """Customize str method."""
        self_name = type(self).__name__
        delivery_type_str = pretty(self.delivery_type().tp)

        return f"{self_name} -> {delivery_type_str}"

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        self_name = type(self).__name__
        delivery_type_str = pretty(self.delivery_type().tp)

        return f"{self_name} will produce a {delivery_type_str}"

    @abstractmethod
    def read(self) -> Self:
        """Encapsulate reading operations."""

    @abstractmethod
    def process(self) -> Self:
        """Encapsulate business rules into data operations."""

    @abstractmethod
    def write(self) -> Self:
        """Persist data on disk."""

    @abstractmethod
    def get(self) -> Optional[T]:
        """Get returns output data to recicycle in other factories."""

    @classmethod
    def delivery_type(cls) -> DeliveryType:
        """Return delivery type."""
        bases = get_original_bases(cls)
        base_factories = [
            base
            for base
            in bases
            if get_origin(base) is Factory
        ]

        if not base_factories:
            raise NotImplementedError("Factory has no type parameter")

        base_factory, *_ = base_factories
        factory_delivery_type, *_ = get_args(base_factory)

        return DeliveryType(factory_delivery_type)

    @property
    def deliverable(self: Self) -> Deliverable:
        """Return the deliverable from this factory."""
        __type = self.delivery_type()

        return (
            Deliverable[__type](  # type: ignore
                payload=self.get(),
                producer=type(self),
                consumers=self.consumers,
                delivery_id=self.delivery_id
            )
        )

    @classmethod
    def expected_deliveries(cls) -> list[Delivery]:
        """Return a list of the expected deliveries for the factory."""
        return [
            delivery.set_consumer(cls)
            for (_, delivery)
            in list(getmembers(cls, lambda _: isinstance(_, Delivery)))
        ]



"""A collection of utilities."""
from .benchmark_result import BenchmarkResult
from .benchmark_modifier import BenchmarkModifier
from .builder import Builder
from .pretty import pretty

__all__ = [
    "BenchmarkModifier",
    "BenchmarkResult",
    "pretty",
    "Builder"
]



"""PySetl expected exceptions."""


class BasePySetlException(Exception):
    """Throw a PySetl Exception."""


class AlreadyExistsException(BasePySetlException):
    """Throw an exception for an already registered object."""


class PySparkException(BasePySetlException):
    """Throw an exception if issue is related with PySpark."""


class InvalidConnectorException(BasePySetlException):
    """Throw an exception if connector does not satisfies requirements."""


class InvalidConfigException(BasePySetlException):
    """Throw an exception if configuration has an error."""


class InvalidDeliveryException(BasePySetlException):
    """Throw an exception if delivery is duplicated or not found."""


class PipelineException(BasePySetlException):
    """Throw an exception if Pipeline fails."""


class BuilderException(BaseException):
    """Throw if builder fails."""


__all__ = [
    "AlreadyExistsException",
    "PySparkException",
    "InvalidConnectorException",
    "InvalidConfigException",
    "PipelineException"
]



"""
SparkEncoding module
"""
from enum import Enum


class SparkEncoding(Enum):
    """
    Available Spark encodings
    """
    US_ASCII = "US-ASCII"
    ISO_8859_1 = "ISO-8859-1"
    UTF_8 = "UTF-8"
    UTF_16BE = "UTF-16BE"
    UTF_16LE = "UTF-16LE"
    UTF_16 = "UTF-16"



"""HasLogger module."""
import logging


class HasLogger:
    """Mixin to provide logging capabilities to a class."""

    def log_info(self, msg: str) -> None:
        """Log info level message."""
        logging.info(msg)

    def log_debug(self, msg: str) -> None:
        """Log debug level message."""
        logging.debug(msg)

    def log_warning(self, msg: str) -> None:
        """Log warning level message."""
        logging.warning(msg)

    def log_error(self, msg: str) -> None:
        """Log error level message."""
        logging.error(msg)



"""
Wrapper for factory type comparison.
"""
from __future__ import annotations
from types import GenericAlias
from typing import get_args, get_origin, _GenericAlias  # type: ignore
from typedspark import DataSet


class DeliveryType:
    """
    A wrapper to easily compare delivery types from factories.
    """
    def __init__(self, tp):
        """
        tp: any Python type representaion:
            - common classes (int, str, MyClass)
            - typing (List[int], Dict[str, float], ...)
            - GenericAlias (list[int], dict[str, int], ...)
            - subclass of typedspark.DataSet (DataSet[Citizen], etc.)
        """
        self.tp = tp if not isinstance(tp, DeliveryType) else tp.tp

    def __eq__(self, other) -> bool:
        other_delivery_type = (
            other
            if isinstance(other, DeliveryType)
            else DeliveryType(other)
        )
        t1 = self.tp
        t2 = other_delivery_type.tp

        t1_alias = self.is_alias(t1)
        t2_alias = self.is_alias(t2)

        # 1) Both types are GenericAlias (e.g. list[int], dict[str, float], ...)
        if t1_alias and t2_alias:
            o1, a1 = get_origin(t1), get_args(t1)
            o2, a2 = get_origin(t2), get_args(t2)

            return (o1 is o2) and (a1 == a2)

        # 2) One type is alias and the other is not.
        if (t1_alias or t2_alias):
            return False

        # 3) Both types are subclasses of DataSet (typedspark)
        if issubclass(t1, DataSet) and issubclass(t2, DataSet):
            schema1 = getattr(t1, "_schema_annotations", None)
            schema2 = getattr(t2, "_schema_annotations", None)

            return schema1 is schema2

        # 4) Basic non parameterized types
        return t1 is t2

    def __repr__(self):
        return repr(self.tp)

    def __str__(self) -> str:
        return str(self.tp)

    def is_alias(self, tp: type):
        """
        Check if type is an alias.
        """
        return isinstance(tp, (GenericAlias, _GenericAlias))



"""Base Config module."""
from __future__ import annotations
from abc import abstractmethod
from functools import reduce
from typing import Any, TypeVar, Generic
from pydantic import BaseModel
from typing_extensions import Self
from pysetl.enums import (
    FileStorage,
)


class BaseConfigModel(BaseModel):
    """BaseConfig validation."""

    storage: FileStorage


class ConfigBuilder:
    """Builder methods for Config class."""

    def __init__(self, cls):
        self.cls = cls

    def from_dict(self, __config: dict) -> Config:
        """Config constructor from a dict."""
        return reduce(
            lambda x, y: x.set(y[0], y[1]),
            __config.items(),
            self.cls()
        )

    def from_object(self, __config: object) -> Config:
        """Config constructor from an object."""
        return self.from_dict({
            x: y
            for (x, y)
            in vars(__config).items()
            if not x.startswith("__")}
        )


T = TypeVar("T", bound=BaseConfigModel)


class Config(Generic[T]):
    """Base Config class."""

    @classmethod
    def builder(cls) -> ConfigBuilder:
        """Create builder."""
        return ConfigBuilder(cls)

    def __init__(self, **kwargs) -> None:
        self.params: dict = kwargs

    def __str__(self) -> str:
        settings_string: str = "\n".join([
            f"  {k}: {v}"
            for k, v
            in self.params.items()
        ])

        return f"PySetl {self.__class__.__name__}:\n{settings_string}"

    def __repr__(self) -> str:
        return str(self)

    def __or__(self, other: Config):
        if not isinstance(other, Config):
            return NotImplemented

        return self.builder().from_dict(self.params | other.params)

    def __contains__(self, __value) -> bool:
        return __value in self.params

    def set(self: Self, key: str, value: Any) -> Self:
        """Set a value to a key in the settings."""
        self.params[key] = value

        return self

    def get(self: Self, key: str) -> Any:
        """Return the value of the key in the settings."""
        return self.params[key]

    def has(self: Self, key: str) -> bool:
        """Check if settings contains a specific value."""
        return key in self.params

    @property
    def config_dict(self: Self) -> dict:
        """Return a validated dictionary with the configuration."""
        return vars(self.config)

    @property
    @abstractmethod
    def config(self: Self) -> T:
        """Expose validated config."""

    @property
    @abstractmethod
    def reader_config(self) -> dict:
        """Returns reader configuration."""

    @property
    @abstractmethod
    def writer_config(self) -> dict:
        """Returns writer configuration."""



"""Deliverable Dispatcher module."""
from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Any, Optional, Union
from typing_extensions import Self
from pysetl.utils.mixins import HasRegistry, HasLogger
from pysetl.utils.exceptions import (
    InvalidDeliveryException
)
from .deliverable import Deliverable
from .stage import Stage
from .delivery import Delivery
from .external import External


if TYPE_CHECKING:  # pragma: no cover
    from .factory import Factory


class Dispatcher(HasRegistry[Deliverable], HasLogger):
    """
    Transfer data between different stages.

    Relies on a Directed Acyclic Graph (DAG). It can:
    - collect a Deliverable from a Factory
    - find the right Deliverable from its pool and send it to a Factory
    """

    def add_deliverable(self: Self, __deliverable: Deliverable) -> Self:
        """
        Add a new Deliverable object into Dispatcher's delivery pool.

        Attention: Dispatcher only guarantee that each deliverable object has
        different UUID. In other words, cannot call twice set_delivery with
        the same argument. However, it doesn't check if multiple deliverables
        have the same data type with the same producer and the same consumer
        """
        self.register(__deliverable)

        return self

    def get_deliverable(
            self: Self,
            available_deliverables: list[Deliverable[Any]],
            consumer: type[Factory],
            producer: Union[type[Factory], type[External]]
            ) -> Optional[Deliverable]:
        """
        Find the Deliverable in pool with the given type.

        @param available_deliverables an array of available deliverable
        """
        if len(available_deliverables) == 0:
            self.log_debug("Can't find a deliverable")
            return None

        if len(available_deliverables) == 1:
            deliverable = available_deliverables[0]
            if (
                deliverable.consumers and
                consumer not in deliverable.consumers
            ):
                raise InvalidDeliveryException(
                    f"Can't find {consumer.name()} in consumer list"
                )
            self.log_debug("Find deliverable")

            return deliverable

        matched_by_producer = [
            d
            for d
            in available_deliverables
            if d.producer == producer
        ]

        if len(matched_by_producer) == 0:
            self.log_warning(
                "Can't find any deliverable that matches the producer"
            )
            return None

        if len(matched_by_producer) == 1:
            self.log_debug("Find deiliverable")
            return matched_by_producer[0]
        self.log_info(
            "Multiple deliverables with same type and same producer"
        )
        self.log_info("Matching by consumer")
        matched_by_consumer = [
            d
            for d
            in matched_by_producer
            if len(d.consumers) != 0 and
            consumer in d.consumers
        ]

        if len(matched_by_consumer) == 0:
            self.log_warning(
                "Can't find any deliverable that matches consumer"
            )
            return None

        if len(matched_by_consumer) == 1:
            self.log_debug("Find deliverable")
            return matched_by_consumer[0]

        raise InvalidDeliveryException("".join([
            "Find multiple deliveries having the same"
            "type, producer and consumer"
        ]))

    def find_deliverable_by(
            self: Self,
            predicate: Callable[[Deliverable], bool]) -> list[Deliverable]:
        """Find all deliverables that passes a given predicate."""
        return [
            v
            for _, v
            in self.get_registry().items()
            if predicate(v)
        ]

    def find_deliverable_by_type(
            self: Self,
            delivery_type: type) -> list[Deliverable]:
        """
        Get all Deliverable by payload type.

        @param delivery_type of data
        """
        return self.find_deliverable_by(
            lambda x: x.is_payload_type(delivery_type)
        )

    def find_deliverable_by_name(
            self: Self,
            name: str) -> list[Deliverable]:
        """Find all deliverables matching payload type name."""
        return self.find_deliverable_by(
            lambda x: x.has_same_payload_type_name(name)
        )

    def collect_factory_deliverable(self: Self, factory: Factory) -> Self:
        """Collect and registers the Deliverable from a Factory."""
        self.add_deliverable(factory.deliverable)

        return self

    def collect_stage_deliverables(self: Self, stage: Stage) -> Self:
        """Collect and registers the Deliverable from a Stage."""
        list(map(self.add_deliverable, stage.deliverable))

        return self

    def dispatch_delivery(
            self: Self,
            consumer: Factory,
            delivery: Delivery) -> None:
        """
        Find delivery by type and id.

        Given a consumer and a delivery, finds a delivery from deliverable pool
        with same type and id.
        """
        self.log_debug(
            f"Look for available {delivery.payload_type} in delivery pool"
        )

        available_deliverables = self.find_deliverable_by(
            lambda _: _.is_payload_type(delivery.payload_type) and
            _.delivery_id == delivery.delivery_id
        )

        deliverable: Optional[Deliverable] = self.get_deliverable(
            available_deliverables,
            consumer=type(consumer),
            producer=delivery.producer
        )

        if deliverable is None:
            raise InvalidDeliveryException(
                f"Cannot find type {delivery.payload_type}"
            )

        delivery.set(deliverable.payload)

    def dispatch(self: Self, consumer: Factory) -> Self:
        """
        Dispath deliverable.

        Dispatch the right deliverable object to the corresponding Delivery
        instances of a factory.
        """
        list(map(
            lambda _: self.dispatch_delivery(consumer, _),
            consumer.expected_deliveries()
        ))

        return self



"""
Provides the ConnectorBuilder for constructing specific connector instances
within the pysetl.connector package.
"""
from typing import Optional
from typing_extensions import Self
from pysetl.utils import Builder
from pysetl.enums import FileStorage
from pysetl.config import FileConfig
from pysetl.utils.exceptions import BuilderException
from .csv_connector import CsvConnector
from .json_connector import JsonConnector
from .parquet_connector import ParquetConnector
from .connector import Connector


class ConnectorBuilder(Builder[Connector]):
    """Build a connector based only on the configuration."""

    def __init__(self: Self, config: FileConfig):
        """Initialize ConnectorBuilder."""
        self.config = config
        self.connector: Optional[Connector] = None

    def build(self) -> Self:
        """Create a connector based on the declared storage."""
        storage = self.config.config.storage

        if storage == FileStorage.CSV:
            self.connector = CsvConnector(self.config)
        elif storage == FileStorage.PARQUET:
            self.connector = ParquetConnector(self.config)
        else:
            self.connector = JsonConnector(self.config)

        return self

    def get(self: Self) -> Connector:
        """Return the right connector based on the configutration."""
        if not self.connector:
            raise BuilderException("Connector is not defined")

        return self.connector



"""HasReader module."""
from abc import abstractmethod, ABC
from pyspark.sql import DataFrameReader


class HasReader(ABC):
    """HasReader mixin."""

    @property
    @abstractmethod
    def _reader(self) -> DataFrameReader:
        """Returns a DataFrameReader."""



"""FileConnectorConfig module."""
from typing import Optional
from typing_extensions import Self
from pyspark.sql.types import StructType
from pydantic import ConfigDict
from pysetl.enums import SaveMode, FileStorage
from .aws_credentials import AwsCredentials
from .config import Config, BaseConfigModel


class FileConfigModel(BaseConfigModel):
    """Validator for file configurations."""

    model_config = ConfigDict(
        extra="allow",
        frozen=True,
        strict=True,
        arbitrary_types_allowed=True
    )

    storage: FileStorage
    path: str
    partition_by: list[str] = []
    savemode: SaveMode = SaveMode.ERRORIFEXISTS
    data_schema: Optional[StructType] = None
    aws_credentials: Optional[AwsCredentials] = None


class FileConfig(Config[FileConfigModel]):
    """Configuration for file connectors."""

    @property
    def config(self: Self) -> FileConfigModel:
        """Returns validated configuration."""
        return FileConfigModel(**self.params)

    @property
    def reader_config(self: Self) -> dict:
        """Configurations passed to DataFrameReader."""
        return {
            k: v
            for k, v
            in self.config
            if k not in FileConfigModel.model_fields.keys()
        }

    @property
    def writer_config(self) -> dict:
        """Configurations passed to DataFrameWriter."""
        return {
            k: v
            for k, v
            in self.config
            if k not in FileConfigModel.model_fields.keys()
        }



"""
utils.mixins module.

Collection of class mixins to add functionalities to base
package classes
"""
from .can_create import CanCreate
from .can_delete import CanDelete
from .can_drop import CanDrop
from .can_partition import CanPartition
from .can_update import CanUpdate
from .can_wait import CanWait
from .has_benchmark import HasBenchmark
from .has_diagram import HasDiagram
from .has_logger import HasLogger
from .has_reader import HasReader
from .has_registry import HasRegistry
from .has_spark_session import HasSparkSession
from .has_writer import HasWriter
from .is_identifiable import IsIdentifiable
from .is_writable import IsWritable
from .is_configurable import IsConfigurable


__all__ = [
    "CanCreate",
    "CanDelete",
    "CanDrop",
    "CanPartition",
    "CanUpdate",
    "CanWait",
    "HasBenchmark",
    "HasDiagram",
    "HasLogger",
    "HasReader",
    "HasRegistry",
    "HasSparkSession",
    "HasWriter",
    "IsConfigurable",
    "IsIdentifiable",
    "IsWritable"
]



"""
Provides the JsonConnector for reading and writing JSON files within the
pysetl.connector package.
"""
from typing import Union
from typing_extensions import Self
from pysetl.config import JsonConfig, FileConfig
from pysetl.utils.exceptions import InvalidConfigException
from pysetl.enums import FileStorage
from .file_connector import FileConnector


class JsonConnector(FileConnector):
    """JsonConnector allows to read JSON files."""

    def __init__(self: Self, options: Union[FileConfig, JsonConfig]) -> None:
        """Initialize JsonConnector."""
        if options.config.storage != FileStorage.JSON:
            raise InvalidConfigException(
                "Not a JsonConfig for a JsonConnector"
            )

        super().__init__(options)



"""Abstract Builder module."""
from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar
from typing_extensions import Self


T_co = TypeVar("T_co", covariant=True)


class Builder(Generic[T_co], metaclass=ABCMeta):
    """Abstract base Builder class."""

    @abstractmethod
    def build(self: Self) -> Self:
        """Build an object."""

    @abstractmethod
    def get(self: Self) -> T_co:
        """Return built object."""

    def getOrCreate(self: Self) -> T_co:  # pylint: disable=C0103
        """Build the object and returns its value."""
        return self.build().get()



"""
S3 Credentials
"""
from pydantic import BaseModel


class AwsCredentials(BaseModel):
    """
    S3 AWS Credentials container
    """
    access_key: str
    secret_key: str
    session_token: str
    credentials_provider: str



"""CanWait module."""
from abc import abstractmethod, ABC
from typing import Optional


class CanWait(ABC):
    """
    CanWait mixin.

    Abstract mixin to add wait functionality to a connector. This methods
    should asynchronously await for result to finish
    """

    @abstractmethod
    def await_termination(self, timeout: Optional[float]) -> None:
        """Await until process finishes."""

    @abstractmethod
    def stop(self) -> None:
        """Stop execution of a process if running."""



"""External module."""
from abc import ABCMeta
from typing import final
from pysetl.utils.mixins import IsIdentifiable


@final
class External(IsIdentifiable, metaclass=ABCMeta):
    """External is the default deliverable producer."""

    @classmethod
    def delivery_type(cls) -> type:
        """Return delivery type of this class."""
        return External



"""PySetl: Python Spark ETL Framework."""
from __future__ import annotations
from uuid import UUID
from typing import Optional, TypeVar
from typing_extensions import Self
from pyspark import SparkConf
from pyspark.sql import SparkSession
from pysetl.config.config import Config
from pysetl.utils.mixins import HasRegistry
from pysetl.utils import Builder
from pysetl.storage import SparkRepositoryBuilder
from pysetl.storage.repository import SparkRepository
from pysetl.utils.exceptions import InvalidConfigException, BuilderException
from pysetl.workflow import Deliverable, Pipeline, Factory


__version__ = "0.1.3b1"


C = TypeVar("C", bound=Config)
ConfigDict = dict[str, C]


class PySetl(HasRegistry[Pipeline]):
    """PySetl is a Spark ETL framework for Python."""

    def __init__(
            self: Self,
            spark: SparkSession,
            config_dict: ConfigDict,
            benchmark: bool = False) -> None:
        """Pysetl Constructor."""
        super().__init__()
        self.spark: SparkSession = spark
        self.external_inputs: dict[str, Deliverable] = {}
        self.config_dict = config_dict
        self.benchmark = benchmark

    def has_external_input(self: Self, external_id: str) -> bool:
        """Check if a deliverable is set in Pipeline dispatcher."""
        return external_id in self.external_inputs

    def list_inputs(self: Self) -> dict[str, Deliverable]:
        """List Deliverable registered in Pipeline Dispatcher."""
        return self.external_inputs

    def get_sparkrepository(
            self: Self,
            repository_type: type,
            config_key: str) -> Optional[SparkRepository]:
        """Get SparkRepository registered in Pipeline Dispatcher."""
        self.set_spark_repository_from_config(repository_type, config_key)

        repository_delivery = self.external_inputs.get(config_key)

        return repository_delivery.payload if repository_delivery else None

    def reset_spark_repository_from_config(
            self: Self,
            repository_type: type,
            config_key: str,
            consumers: Optional[set[type[Factory]]] = None,
            delivery_id: str = "",
            cache: bool = False) -> Self:
        """Create a SparkRepository and register it in Dispatcher."""
        _consumers = consumers if consumers is not None else set()
        config = self.config_dict.get(config_key)

        if not config:
            raise InvalidConfigException(
                f"No config found with key: {config_key}"
            )

        repository = SparkRepositoryBuilder[repository_type](  # type: ignore
            config,
            cache=cache
        ).getOrCreate()

        repository_deliverable = (
            Deliverable[SparkRepository[repository_type]](  # type: ignore
                repository
            )
            .set_consumers(_consumers)
            .set_delivery_id(delivery_id)
        )

        self.external_inputs[config_key] = repository_deliverable

        return self

    def set_spark_repository_from_config(
            self: Self,
            repository_type: type,
            config_key: str,
            consumers: Optional[set[type[Factory]]] = None,
            delivery_id: str = "",
            cache: bool = False) -> Self:
        """Check if SparkRepository already registered otherwise reset it."""
        if not self.has_external_input(config_key):
            self.reset_spark_repository_from_config(
                repository_type,
                config_key,
                consumers,
                delivery_id,
                cache
            )

        return self

    def reset_spark_repository(
            self: Self,
            repository: SparkRepository,
            repository_id: str,
            consumers: Optional[set[type[Factory]]] = None,
            delivery_id: str = "") -> Self:
        """Register SparkRepository in Pipeline Dispatcher."""
        _consumers = consumers if consumers else set()
        __type = repository.type_annotation
        repository_deliverable = Deliverable[__type](  # type: ignore
            repository,
            consumers=_consumers,
            delivery_id=delivery_id
        )
        self.external_inputs[repository_id] = repository_deliverable

        return self

    def set_spark_repository(
            self: Self,
            repository: SparkRepository,
            repository_id: str,
            consumers: Optional[set[type[Factory]]] = None,
            delivery_id: str = "") -> Self:
        """Check if SparkRepositoty registered otherwise, register it."""
        if not self.has_external_input(repository_id):
            self.reset_spark_repository(
                repository,
                repository_id,
                consumers,
                delivery_id
            )

        return self

    def new_pipeline(self: Self) -> Pipeline:
        """Create Pipeline and register it in PySel object."""
        pipeline = Pipeline(benchmark=self.benchmark)

        self.register(pipeline)

        list(map(
            lambda t: pipeline.set_input_from_deliverable(t[1]),
            self.external_inputs.items()
        ))

        return pipeline

    def get_pipeline(self: Self, uuid: UUID) -> Optional[Pipeline]:
        """Get Pipeline from registry."""
        return self.get_registry().get(uuid)

    def stop(self: Self) -> None:
        """Stop SparkSession."""
        self.spark.stop()

    @classmethod
    def builder(cls) -> PySetlBuilder:
        """Return a builder for PySetl."""
        return PySetlBuilder()


class PySetlBuilder(Builder[PySetl]):
    """Build a PySetl object."""

    def __init__(
            self: Self,
            benchmark: bool = False,
            config: Optional[ConfigDict] = None,
            shuffle_partitions: Optional[int] = None,
            spark_master: str = "local",
            spark_session: Optional[SparkSession] = None,
            spark_config: Optional[SparkConf] = None) -> None:
        self.benchmark = benchmark
        self.config = config
        self.spark_config = spark_config
        self.shuffle_partitions = shuffle_partitions
        self.spark_master = spark_master
        self.spark_session = spark_session
        self.__pysetl: Optional[PySetl] = None

    def set_config(self: Self, config: ConfigDict) -> Self:
        """Set PySetl SparkRepository configurations."""
        self.config = config

        return self

    def set_spark_config(self: Self, spark_config: SparkConf) -> Self:
        """Set SparkSession configuration."""
        self.spark_config = spark_config

        return self

    def set_shuffle_partitions(self: Self, shuffle: int) -> Self:
        """Set SparkSession number of shuffle partitions."""
        self.shuffle_partitions = shuffle

        return self

    def set_spark_master(self: Self, master: str) -> Self:
        """Set SparkSession master."""
        self.spark_master = master

        return self

    def set_spark_session(self: Self, session: SparkSession):
        """Set SparkSession directly."""
        self.spark_session = session

        return self

    def build_spark_session(self: Self) -> SparkSession:
        """Build SparkSession."""
        shuffle_partitions = (
            [(
                str("spark.sql.shuffle.partitions"),
                str(self.shuffle_partitions)
            )]
            if self.shuffle_partitions
            else []
        )
        config_from_dict: Optional[Config] = (
            self.config.get("spark")
            if self.config
            else None
        )
        spark_app_name = config_from_dict.params.get(
            "app_name",
            "default_app_name"
        ) if config_from_dict else "default"
        spark_config_dict = list(
            config_from_dict.params.items()
        ) if config_from_dict else []
        spark_config_from_builder = (
            self.spark_config.getAll()
            if self.spark_config
            else []
        )
        final_config: SparkConf = (
            SparkConf()
            .setAll(spark_config_from_builder)
            .setAll(spark_config_dict)
            .setAll(shuffle_partitions)
        )

        session_builder: SparkSession.Builder = (
            SparkSession.Builder()
            .master(self.spark_master)
            .appName(spark_app_name)
            .config(conf=final_config)
        )

        return session_builder.getOrCreate()

    def build(self: Self) -> Self:
        """Build PySetl."""
        spark = self.spark_session \
            if self.spark_session \
            else self.build_spark_session()

        config = self.config if self.config else {}

        self.__pysetl = PySetl(spark, config, self.benchmark)

        return self

    def get(self: Self) -> PySetl:
        """Return PySetl."""
        if not self.__pysetl:
            raise BuilderException("No PySetl object built")

        return self.__pysetl



"""HasRegistryMixin module."""
from typing import TypeVar, Generic, Optional
from uuid import UUID
from typing_extensions import Self
from pysetl.utils.exceptions import AlreadyExistsException
from pysetl.utils.mixins.is_identifiable import IsIdentifiable


T = TypeVar("T", bound=IsIdentifiable)


class HasRegistry(Generic[T]):
    """Mixin class to give child a registry of unique objects."""

    def __init__(self) -> None:
        self._registry: dict[UUID, T] = {}

    def is_registered(self, item: T) -> bool:
        """Check if item is registered."""
        return item.uuid in self._registry

    def clear_registry(self):
        """Restore registry."""
        self._registry = {}

        return self

    def register_items(self, items: list[T]):
        """Register multiple items."""
        for item in items:
            self.register(item)

        return self

    def register(self, item: T):
        """Register an item."""
        if self.is_registered(item):
            raise AlreadyExistsException(
                f"The item {item.uuid} of type {item.name()} already exists"
            )

        self._registry[item.uuid] = item

        return self

    def get_registry(self) -> dict[UUID, T]:
        """Expose the registry."""
        return self._registry

    def get_item(self, uuid: UUID) -> Optional[T]:
        """Get an item by its UUID."""
        return self._registry.get(uuid, None)

    @property
    def last_registered_item(self: Self) -> Optional[T]:
        """Return last registered item."""
        if len(self._registry) == 0:
            return None

        return list(self._registry.items())[-1][1]

    @property
    def size(self: Self) -> int:
        """Return dize of the registry."""
        return len(self._registry)



"""FileStorage module."""
from enum import Enum


class FileStorage(Enum):
    """Available general file storage options."""

    CSV = "csv"
    PARQUET = "parquet"
    JSON = "json"
    GENERIC = "generic"



"""DAG module."""
from __future__ import annotations
from dataclasses import dataclass
from typing_extensions import Self, TYPE_CHECKING
from pysetl.utils.mixins import HasDiagram
from .flow import Flow
from .node import Node
from .delivery import Delivery


if TYPE_CHECKING:
    from .factory import Factory  # pragma: no cover


@dataclass
class DAG(HasDiagram):
    """
    A Directed Acyclig Graph.

    A DAG is a way to represent the flow of information in a process. It should
    be acyclic and direct in order to ensure that information flow will solve
    correctly. Each node in the graph represents a Factory and each Factory
    returns a Deliverable. Factories can recycle information from parent nodes
    and dispatch the Deliverable as a Delivery to child nodes.
    """

    nodes: set[Node]
    flows: set[Flow]

    def find_deliveries(self: Self, factory: Factory) -> list[Delivery]:
        """Find deliveries from a given Factory in the nodes of the DAG."""
        return [
            delivery
            for node
            in self.nodes
            for delivery
            in node.expected_deliveries
            if node.factory_uuid == factory.uuid
        ]

    def to_diagram(self: Self) -> str:
        """Create mermaid diagram for DAG."""
        node_diagrams = "\n".join([n.to_diagram() for n in self.nodes])
        flow_diagrams = "\n".join([f.to_diagram() for f in self.flows])

        return "\n".join([
            "classDiagram",
            node_diagrams,
            flow_diagrams
        ])

    @property
    def diagram_id(self: Self) -> str:
        """Raise an error beacause it is not implemented."""
        raise NotImplementedError("DAG doesn't have diagram id")

    def __str__(self: Self) -> str:
        """Customize str method."""
        empty_nodes_string = "Empty\n" if len(self.nodes) == 0 else ""
        empty_flows_string = "Empty\n" if len(self.flows) == 0 else ""
        nodes_string = "\n".join([
            str(n)
            for n
            in sorted(list(self.nodes), key=lambda x: x.stage)
        ])
        flows_string = "\n".join([
            str(f)
            for f
            in sorted(list(self.flows), key=lambda x: x.stage)
        ])

        return "\n".join([
            "-------------   Data Transformation Summary  -------------",
            empty_nodes_string,
            nodes_string,
            "------------------   Data Flow Summary  ------------------",
            empty_flows_string,
            flows_string,
        ])

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        return self.__str__()



"""CanDrop module."""
from abc import abstractmethod, ABC


class CanDrop(ABC):
    """Abstract Mixin to add delete functionality to a connector."""

    @abstractmethod
    def drop(self) -> None:
        """Drop a table from de storage."""



"""HasSparkSession module."""
from pyspark.sql import SparkSession


class HasSparkSession:
    """Provide current spark session property."""

    @property
    def spark(self):
        """Returns current spark session."""
        return SparkSession.Builder().getOrCreate()



"""NodeOutput module."""
from __future__ import annotations
from dataclasses import dataclass, is_dataclass, fields
from typing import (  # type: ignore
    TYPE_CHECKING, get_origin, get_args, TypeVar,
    _GenericAlias  # type: ignore
)
from typing_extensions import Self
from typedspark import DataSet
from pydantic import BaseModel
from pysetl.utils.mixins import HasDiagram
from pysetl.utils.pretty import pretty
from pysetl.utils.exceptions import BasePySetlException
from pysetl.workflow.delivery_type import DeliveryType


if TYPE_CHECKING:  # pragma: no cover
    from .factory import Factory


@dataclass
class NodeOutput(HasDiagram):
    """Node output representation."""

    delivery_type: type
    consumer: set[type[Factory]]
    delivery_id: str = ""
    final_output: bool = False
    external: bool = False

    @property
    def wrapped_delivery_type(self) -> DeliveryType:
        """Wraps delivery type for easy comparison."""
        return DeliveryType(self.delivery_type)

    @property
    def diagram_id(self) -> str:
        """Digram identifier."""
        final_suffix = "Final" if self.final_output else ""
        external_suffix = "External" if self.external else ""

        return self.format_diagram_id(
            pretty(self.delivery_type),
            self.delivery_id,
            final_suffix + external_suffix
        )

    def get_fields(self: Self) -> list[str]:
        """Get arguments from payload signature."""
        def _pretty(arg, tpe):
            return pretty(arg) if isinstance(tpe, TypeVar) else pretty(tpe)

        origin = get_origin(self.delivery_type)
        args = get_args(self.delivery_type)

        if (
            isinstance(self.delivery_type, _GenericAlias) and
            is_dataclass(origin)
        ):
            (arg,) = args

            return [
                f"    >{field.name}: {_pretty(arg, field.type)}"
                for field
                in fields(origin)
            ]

        if issubclass(self.delivery_type, DataSet):
            schema = getattr(self.delivery_type, "_schema_annotations", None)

            if not schema:
                raise BasePySetlException("DataSet has no schema.")

            return [
                f"    >{field.name}: {field.dataType.simpleString()}"
                for field
                in schema.get_structtype().fields
            ]

        if is_dataclass(self.delivery_type):
            return [
                f"    >{field.name}: {pretty(field.type)}"
                for field
                in fields(self.delivery_type)
            ]

        if issubclass(self.delivery_type, BaseModel):
            return [
                f"    >{key}: {pretty(value.annotation)}"
                for key, value
                in self.delivery_type.model_fields.items()
            ]

        return []

    def to_diagram(self) -> str:
        """Create mermaid diagram for a NodeOutput."""
        fields_str: str = "\n".join(self.get_fields())

        return "\n".join([
            f"class {self.diagram_id} {{",
            f"    <<{pretty(self.delivery_type)}>>",
            fields_str,
            "}"
        ])



"""
The pysetl.storage.repository package. Provides high-level, type-safe data
access layers (repositories) over connectors.
"""
from .base_repository import BaseRepository
from .sparkrepository import SparkRepository
from .sparkrepository_builder import SparkRepositoryBuilder


__all__ = [
    "BaseRepository",
    "SparkRepository",
    "SparkRepositoryBuilder"
]



"""HasDiagram module."""

from abc import abstractmethod, ABC
from inspect import signature
from typing_extensions import Self


class HasDiagram(ABC):
    """Mixin for objects that can be represented as a mermaid diagram."""

    @property
    @abstractmethod
    def diagram_id(self: Self) -> str:
        """Return diagram id."""

    @abstractmethod
    def to_diagram(self: Self) -> str:
        """Generate the diagram."""

    def get_signature(self):
        """Return constructor signature."""
        return list(signature(self.__init__).parameters)

    def format_diagram_id(
            self: Self,
            name: str,
            diagram_id: str,
            suffix: str) -> str:
        """Return formatted diagram ID."""
        return (
            name.replace("[", "").replace("]", "") +
            diagram_id.capitalize() +
            suffix
        )



"""
Provides the SparkRepositoryBuilder for constructing SparkRepository instances
within the pysetl.storage.repository package.
"""
from typing import TypeVar, Generic, Optional, get_args
from typing_extensions import Self
from typedspark import Schema
from pysetl.utils import Builder
from pysetl.utils.exceptions import BuilderException
from pysetl.config import FileConfig
from pysetl.storage.connector import Connector, ConnectorBuilder
from .sparkrepository import SparkRepository


T = TypeVar("T", bound=Schema)


class SparkRepositoryBuilder(Generic[T], Builder[SparkRepository[T]]):
    """
    Builder for SparkRepository.

    The SparkRepositoryBuilder will build a SparkRepository according to the
    given DataType and Storage
    @param storage type of storage
    @param A Config object
    """

    def __init__(
            self: Self,
            config: FileConfig,
            cache: bool = False) -> None:
        """Initialize builder."""
        self.config = config
        self.cache = cache
        self.__sparkrepository: Optional[SparkRepository[T]] = None

    def __create_connector(self: Self) -> Connector:
        """Create the connector for the repository."""
        return ConnectorBuilder(self.config).getOrCreate()

    def build(self: Self) -> Self:
        """Build repository from config."""
        connector = self.__create_connector()
        orig_class = getattr(self, "__orig_class__")

        [__type, *_] = get_args(orig_class)  # pylint: disable=E1101
        self.__sparkrepository = SparkRepository[__type](  # type: ignore
            connector,
            self.cache
        )

        return self

    def get(self: Self) -> SparkRepository[T]:
        """Return built repository."""
        if not self.__sparkrepository:
            raise BuilderException("No SparkRepository built")
        return self.__sparkrepository



"""
Provides the ParquetConnector for reading and writing Parquet files within the
pysetl.connector package.
"""
from typing import Union
from typing_extensions import Self
from pysetl.config import ParquetConfig, FileConfig
from pysetl.utils.exceptions import InvalidConfigException
from pysetl.enums import FileStorage
from .file_connector import FileConnector


class ParquetConnector(FileConnector):
    """ParquetConnector allows to read PARQUET files."""

    def __init__(
            self: Self,
            options: Union[FileConfig, ParquetConfig]) -> None:
        """Initialize ParquetConnector."""
        if options.config.storage != FileStorage.PARQUET:
            raise InvalidConfigException(
                "Not a ParquetConfig for a ParquetConnector"
            )

        super().__init__(options)



"""
DbStorage module
"""
from enum import Enum


class DbStorage(Enum):
    """
    Available general DB storage options
    """
    DELTA = "DELTA"
    CASSANDRA = "CASSANDRA"
    DYNAMODB = "DYNAMODB"
    HUDI = "HUDI"



"""Workflow module."""
from .pipeline import Pipeline
from .stage import Stage
from .deliverable import Deliverable
from .delivery import Delivery
from .factory import Factory


__all__ = [
    "Pipeline",
    "Stage",
    "Deliverable",
    "Factory",
    "Delivery"
]



"""
Defines the abstract Connector class, integrating Spark and logging, for the
pysetl.connector package.
"""
from abc import ABCMeta
from pysetl.utils.mixins import HasSparkSession, HasLogger
from pysetl.config.config import BaseConfigModel
from .base_connector import BaseConnector


class Connector(BaseConnector, HasSparkSession, HasLogger, metaclass=ABCMeta):
    """
    Abstract Connector.

    A connector should implement read/write operations with a SparkSession
    object. These implementations should be architecture specific
    """

    config: BaseConfigModel



"""Writable Mixin."""
from typing_extensions import Self


class IsWritable:
    """Indicate that users can activate or deactivate write of the class."""

    _write: bool = True

    @property
    def is_writable(self) -> bool:
        """Return true if the write method will be invoked by the pipeline."""
        return self._write

    def writable(self, write: bool) -> Self:
        """
        Whether invoke the write method or not.

        @param write if set to true, then the write method of the factory will
        be invoked
        """
        self._write = write

        return self



"""
Provides the CsvConnector for reading and writing CSV files within the
pysetl.connector package.
"""
from typing import Union
from typing_extensions import Self
from pysetl.config import CsvConfig, FileConfig
from pysetl.utils.exceptions import InvalidConfigException
from pysetl.enums import FileStorage
from .file_connector import FileConnector


class CsvConnector(FileConnector):
    """CsvConnector allows to read CSV files."""

    def __init__(self: Self, options: Union[FileConfig, CsvConfig]) -> None:
        """Initialize CsvConnector."""
        if options.config.storage != FileStorage.CSV:
            raise InvalidConfigException("Not a CsvConfig for a CsvConnector")

        super().__init__(options)



"""Delivery module."""
from __future__ import annotations
from typing import TYPE_CHECKING, Generic, TypeVar, Optional, Union, get_args
from typing_extensions import Self
from pysetl.utils.exceptions import InvalidDeliveryException
from pysetl.utils import pretty
from .external import External


if TYPE_CHECKING:  # pragma: no cover
    from .factory import Factory


T = TypeVar("T")


class Delivery(Generic[T]):
    """
    Container for results from another factory.

    Class designed to retrive data from the Deliverable pool
    """

    def __init__(
        self: Self,
        producer: type[Union[Factory, External]] = External,
        delivery_id: str = "",
        consumer: Optional[type[Factory]] = None
    ) -> None:
        self.producer = producer
        self.delivery_id = delivery_id
        self.consumer = consumer
        self.__value: Optional[T] = None

    def __str__(self: Self) -> str:
        """Customize str method."""
        _producer = pretty(self.producer)
        _type = pretty(self.payload_type)

        return f"Delivery from {_producer} with type {_type}"

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        return str(self)

    def set(self: Self, __value: T) -> Self:
        """Set the value of the delivery."""
        self.__value = __value

        return self

    def get(self: Self) -> T:
        """Get delivery."""
        if not self.__value:
            raise InvalidDeliveryException("Delivery with no value")

        return self.__value

    @property
    def payload_type(self: Self) -> type:
        """Return payload type."""
        instance_type = getattr(self, "__orig_class__")
        (type_param,) = get_args(instance_type)

        return type_param

    def set_consumer(self: Self, __value: type[Factory]) -> Self:
        """Set Delivery consumer."""
        self.consumer = __value

        return self



"""
BenchmarkModifier module
"""
import inspect
import types
import time
from typing import Generic, TypeVar, Callable
from typing_extensions import Self


T = TypeVar("T")


class BenchmarkModifier(Generic[T]):
    """
    Wrapper for any function to benchmark its method calls
    """
    def __init__(self: Self, obj: T) -> None:
        self._obj = obj
        for name, value in inspect.getmembers(obj, callable):
            if not hasattr(self, name) and isinstance(value, types.MethodType):
                setattr(self._obj, name, self.__benchmark(value))

        object.__setattr__(self._obj, "times", {})

    def __benchmark(self: Self, function: Callable) -> Callable:
        """
        Benchmark a function execution time
        """
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = function(*args, **kwargs)
            elapsed_time = time.time() - start_time
            self._obj.times[function.__name__] = elapsed_time

            return result

        return wrapper

    def get(self) -> T:
        """
        Returns modified object
        """
        return self._obj



"""Node class."""
from __future__ import annotations
from dataclasses import dataclass
from uuid import UUID
from typing import TYPE_CHECKING, Union
from typing_extensions import Self
from pysetl.utils.mixins import (
    HasDiagram, HasLogger,
    IsIdentifiable
)
from pysetl.utils.exceptions import InvalidDeliveryException
from pysetl.utils import pretty
from .delivery import Delivery
from .node_output import NodeOutput
from .external import External


if TYPE_CHECKING:  # pragma: no cover
    from .factory import Factory


@dataclass
class Node(IsIdentifiable, HasDiagram, HasLogger):
    """
    Node is a representation of Factory in the DAG.

    One node could have multiple inputs and one single output.
    """

    factory_class: Union[type[Factory], type[External]]
    factory_uuid: UUID
    stage: int
    expected_deliveries: list[Delivery]
    output: NodeOutput

    def __post_init__(self: Self) -> None:
        """Initialize UUID from IsIdentifiable."""
        super().__init__()

    def __hash__(self: Self) -> int:
        """Hash for set operations."""
        return self.uuid.int

    @classmethod
    def from_factory(
            cls,
            factory: Factory,
            stage: int,
            final: bool) -> Node:
        """Create a node from a Factory."""
        return Node(
            factory_class=type(factory),
            factory_uuid=factory.uuid,
            stage=stage,
            expected_deliveries=factory.expected_deliveries(),
            output=NodeOutput(
                delivery_type=factory.delivery_type().tp,
                consumer=factory.consumers,
                delivery_id=factory.delivery_id,
                final_output=final
            )
        )

    @property
    def input(self) -> list[Delivery]:
        """
        List of NodeInputs.

        This should be metadata of input factories.
        """
        return self.expected_deliveries

    @property
    def diagram_id(self: Self) -> str:
        """Return string with diagram id."""
        return pretty(self.factory_class).replace("[]", "_")

    @property
    def producers(self: Self) -> list[Union[type[Factory], type[External]]]:
        """Return a list of input producer factories."""
        return [delivery.producer for delivery in self.input]

    def find_input_by_type(
            self,
            payload_type: type,
            delivery_id: str) -> list[Delivery]:
        """Search inside input nodes for a given factory class."""
        return [
            delivery
            for delivery
            in self.input
            if delivery.producer is not None and
            payload_type == delivery.payload_type and
            delivery_id == delivery.delivery_id
        ]

    def __format_delivery_id(self: Self, delivery_id) -> str:
        """Format Delivery ID to print in diagram."""
        return f" (delivery id: {delivery_id})" if delivery_id != "" else ""

    def __format_delivery(self: Self, delivery: Delivery) -> str:
        delivery_type_str = (
            pretty(delivery.producer.delivery_type())
            if delivery.producer
            else "External"
        )
        delivery_id_str = self.__format_delivery_id(delivery.delivery_id)

        return f"Input: {delivery_type_str}{delivery_id_str}"

    def __str__(self) -> str:
        """Describe the node contents."""
        node_name = pretty(self.factory_class)
        formated_inputs = [
            self.__format_delivery(delivery)
            for delivery
            in self.input
        ]
        input_lines = "\n".join(formated_inputs)
        output_type = pretty(self.output.delivery_type)
        output_delivery_id = self.__format_delivery_id(
            self.output.delivery_id
        )
        output_line = f"Output: {output_type} {output_delivery_id}"

        return "\n".join([
            f"Factory: {node_name}",
            f"Stage: {self.stage}",
            input_lines,
            output_line,
            "----------------------------------------------------------"
        ])

    def __repr__(self: Self) -> str:
        """Cutomize repr method."""
        return str(self)

    def to_diagram(self) -> str:
        """Create mermaid diagram."""
        input_types = map(
            lambda i: f"    +{pretty(i.payload_type)}",
            self.input
        )
        input_types_str = "\n".join(input_types)

        this_diagram_string = "\n".join([
            f"class {self.diagram_id} {{",
            f"     <<Factory[{pretty(self.output.delivery_type)}]>>",
            input_types_str,
            "}"
        ])

        output_str = self.output.to_diagram()
        link = f"{self.output.diagram_id} <|.. {self.diagram_id} : Output"

        return f"{this_diagram_string}\n{output_str}\n{link}"

    def target_node(self: Self, next_node: Node) -> bool:
        """
        Return true if it is a target node of the current node.

        False otherwise.
        """
        valid_node_uuid = self.uuid != next_node.uuid
        valid_class_uuid = self.factory_uuid != next_node.factory_uuid
        valid_stage = self.stage < next_node.stage

        filtered_inputs: list[Delivery] = next_node.find_input_by_type(
            self.output.delivery_type,
            self.output.delivery_id
        )

        if len(filtered_inputs) == 0:
            valid_target = False
        elif len(filtered_inputs) == 1:
            valid_target = self.__handle_one_matches(
                filtered_inputs[0],
                next_node
            )
        else:
            valid_target = self.__handle_multiple_matches(
                filtered_inputs,
                next_node
            )

        return (
            valid_node_uuid and
            valid_class_uuid and
            valid_stage and
            valid_target
        )

    def __handle_one_matches(
            self: Self,
            delivery: Delivery,
            next_node: Node) -> bool:
        """Define helper to validate target when one delivery matches."""
        valid_producer = (
            delivery.producer is External or
            delivery.producer is self.factory_class
        )
        valid_consumer = (
            len(self.output.consumer) == 0 or
            next_node.factory_class in self.output.consumer
        )

        return valid_producer and valid_consumer

    def __handle_multiple_matches(
            self: Self,
            deliveries: list[Delivery],
            next_node: Node) -> bool:
        """Define helper to validate target when multiple deliveries match."""
        valid_consumer = (
            len(self.output.consumer) == 0 or
            next_node.factory_class in self.output.consumer
        )

        exact_produced_match = any(
            delivery.producer == self.factory_class
            for delivery
            in deliveries
        )
        non_explicitly_defined_producers = [
            delivery
            for delivery
            in deliveries
            if delivery.producer is External
        ]

        if (
            not exact_produced_match and
            len(non_explicitly_defined_producers) > 1
        ):
            raise InvalidDeliveryException(f"""
            Multiple inputs in {pretty(type(next_node))} are of type
            {self.output.delivery_type}. You may have to explicitly declare
            their producer in the Delivery annotation, otherwise this may cause
            unexpected pipeline result.
            """)

        valid_producer = (
            exact_produced_match or
            len(non_explicitly_defined_producers) == 1
        )

        return valid_consumer and valid_producer

    @classmethod
    def external(cls, output: NodeOutput) -> Node:
        """Create an external source node."""
        return Node(
            External,
            UUID("{00000000-0000-0000-0000-000000000000}"),
            -1,
            [],
            output
        )

    def find_input_by_delivery_id(
            self: Self,
            delivery_id: str) -> list[Delivery]:
        """Search inside input nodes for a given delivery id."""
        return [
            delivery
            for delivery
            in self.input
            if delivery.delivery_id == delivery_id
        ]



"""
SaveMode module
"""
from enum import Enum


class SaveMode(Enum):
    """
    Available Spark SaveModes
    """
    APPEND = "append"
    ERRORIFEXISTS = "errorifexists"
    IGNORE = "ignore"
    OVERWRITE = "overwrite"



"""Stage module."""
from __future__ import annotations
from typing import Callable, Any, Optional
from functools import reduce
from typing_extensions import Self, ParamSpec
from pyspark.sql import SparkSession
from pysetl.utils.mixins import (
    HasLogger, IsIdentifiable, HasRegistry, HasBenchmark,
    IsWritable
)
from pysetl.utils import BenchmarkResult, BenchmarkModifier
from .deliverable import Deliverable
from .node import Node
from .factory import Factory


P = ParamSpec("P")


class Stage(
    HasRegistry[Factory],
    HasLogger,
    HasBenchmark,
    IsWritable,
    IsIdentifiable
):
    """
    A Stage is a collection of independent Factories.

    All the stages of a pipeline will be executed sequentially at runtime.
    Within a stage, all factories are executed in sequential order.
    """

    def __init__(self, benckmark: bool = False) -> None:
        super().__init__()
        IsIdentifiable.__init__(self)
        self.benchmarked = benckmark
        self.__end: bool = True
        self.__stage_id: int = 0
        self.__deliverable: list[Deliverable] = []
        self.__benchmark_result: list[BenchmarkResult] = []

    @property
    def end(self: Self) -> bool:
        """Exposes private end property."""
        return self.__end

    @end.setter
    def end(self: Self, __value: bool) -> None:
        """Exposes private end property."""
        self.__end = __value

    @property
    def start(self: Self) -> bool:
        """Verifies if stage starts DAG."""
        return self.__stage_id == 0

    @property
    def stage_id(self: Self) -> int:
        """Exposes private stage_id property."""
        return self.__stage_id

    @stage_id.setter
    def stage_id(self: Self, __value: int) -> None:
        self.__stage_id = __value

    @property
    def factories(self: Self) -> list[Factory]:
        """Exposes registered factories in the stage."""
        return list(self.get_registry().values())

    @property
    def deliverable(self: Self) -> list[Deliverable]:
        """Returns the list of deliverables of the stage."""
        return self.__deliverable

    def instantiate_factory(
            self: Self,
            factory_type: type[Factory],
            *args,
            **kwargs) -> Factory:
        """Given a type[Factory] instantiates the class."""
        return factory_type(*args, **kwargs)

    def add_factory_from_type(
            self: Self,
            factory_type: type[Factory],
            *args,
            **kwargs) -> Self:
        """Add a factory to the stage given a type[Factory]."""
        factory = self.instantiate_factory(factory_type, *args, **kwargs)

        return self.add_factory(factory)

    def add_factory(self: Self, factory: Factory):
        """Add a factory to the stage given a Factory instance."""
        self.register(factory)

        return self

    def __str__(self: Self) -> str:
        """Customize str method."""
        self.log_info(
            f"Stage {self.stage_id} contains {len(self.get_registry())} items"
        )
        factories_string = "\n".join([str(f) for f in self.factories])

        return factories_string

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        return str(self)

    def should_write(self: Self, factory: Factory) -> bool:
        """Verify if stage and factory should write."""
        return self.is_writable and factory.is_writable

    def execute_factory(self: Self, factory: Factory) -> Factory:
        """Execute a factory."""
        if self.should_write(factory):
            self.log_debug(f"Persist output of {factory.__class__.__name__}")

            return factory.read().process().write()

        return factory.read().process()

    def handle_benchmark(self, factory: Factory) -> BenchmarkResult:
        """Execute and benchmarks a given Factory."""
        factory_class_name = factory.__class__.__name__

        benckmarked_factory = BenchmarkModifier(factory).get()
        executed_factory = self.execute_factory(benckmarked_factory)

        benchmark_times: dict = executed_factory.times  # type: ignore
        elapsed_time = reduce(lambda x, y: x + y, benchmark_times.values())

        return BenchmarkResult(
            factory_class_name,
            benchmark_times.get("read", 0),
            benchmark_times.get("process", 0),
            benchmark_times.get("write", 0),
            elapsed_time
        )

    def with_spark_session(
        self: Self,
        fun: Callable[[SparkSession], Any]
    ) -> Any:
        """Run a functions if SparkSession exists otherwise returns None."""
        spark: Optional[SparkSession] = SparkSession.getActiveSession()

        if spark:
            return fun(spark)

        return None

    def run_factory(self: Self, factory: Factory) -> Deliverable:
        """Run a factory and collects its deliverable."""
        self.with_spark_session(lambda _: _.sparkContext.setJobGroup(
            factory.__class__.__name__,
            "stage group"
        ))

        if self.benchmarked:
            benchmark_result = self.handle_benchmark(factory)

            self.__benchmark_result.append(benchmark_result)
        else:
            self.execute_factory(factory)

        # Find a way clearJobGroup since sparkContext method doesn't exists
        # self.with_spark_session(lambda _: _.sparkContext.clearJobGroup())
        return factory.deliverable

    def create_nodes(self: Self) -> list[Node]:
        """Create stage nodes with registered stage factories."""
        return [
            Node.from_factory(f, self.__stage_id, self.__end)
            for f
            in self.factories
        ]

    def get_benchmarks(self: Self) -> list[BenchmarkResult]:
        """Return collected benchmarks."""
        return self.__benchmark_result

    def run(self: Self) -> Self:
        """Run the Stage."""
        self.__deliverable = list(map(self.run_factory, self.factories))

        return self



"""BanchmarkResult module."""
from dataclasses import dataclass


@dataclass
class BenchmarkResult:
    """A simple formmatter for factories execution benchmark."""
    cls: str
    read: float
    process: float
    write: float
    total: float

    def __str__(self) -> str:
        return "\n".join([
            f"Benchmark class: {self.cls}",
            f"Total elapsed time: {self.total} s",
            f"read: {self.read} s",
            f"process: {self.process} s",
            f"write: {self.write} s",
            "================="
        ])



"""Flow module."""
from __future__ import annotations
import textwrap
from dataclasses import dataclass
from typing_extensions import Self
from pysetl.utils.mixins import HasDiagram, IsIdentifiable
from pysetl.utils import pretty
from .node import Node


@dataclass
class Flow(IsIdentifiable, HasDiagram):
    """
    Flow is a representation of the data transfer in a Pipeline.

    @param from origin node of the transfer
    @param to   destination node of the transfer
    """

    from_node: Node
    to_node: Node

    def __post_init__(self: Self) -> None:
        IsIdentifiable.__init__(self)

    def __hash__(self: Self) -> int:
        return self.uuid.int

    @property
    def stage(self: Self) -> int:
        """Expose stage from NodeInput."""
        return self.from_node.stage

    @property
    def delivery_id(self: Self) -> str:
        """Expose delivery_id from NodeInput."""
        return self.from_node.output.delivery_id

    def to_diagram(self: Self) -> str:
        """Create mermaid diagram."""
        return (
            f"{self.to_node.diagram_id}<|--" +
            f"{self.from_node.output.diagram_id}:Input"
        )

    @property
    def diagram_id(self: Self) -> str:
        """Flow has no diagram id."""
        return ""

    def __str__(self: Self) -> str:
        """Customize str method."""
        from_factory = pretty(self.from_node.factory_class)
        to_factory = pretty(self.to_node.factory_class)

        return f"Flow[{from_factory} -> {to_factory}]"

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        delivery_id_string = (
            f"Delivery id : {self.delivery_id}"
            if self.delivery_id != ""
            else ""
        )
        from_factory = pretty(self.from_node.factory_class)
        to_factory = pretty(self.to_node.factory_class)

        return textwrap.dedent(f"""
        {delivery_id_string}
        Stage       : {self.stage}
        Direction   : {from_factory} ==> {to_factory}
        PayLoad     : {pretty(self.payload)}
        ----------------------------------------------------------
        """).strip()

    @property
    def payload(self: Self) -> type:
        """Expose payload output from parent node."""
        return self.from_node.output.delivery_type



"""HasWriter module."""
from abc import abstractmethod, ABC
from pyspark.sql import DataFrame, DataFrameWriter


class HasWriter(ABC):
    """HasWrite Mixin."""

    @abstractmethod
    def _writer(self, data: DataFrame) -> DataFrameWriter:
        """Return a DataFrameWriter."""



"""
The pysetl.connector package. Provides connectors for various data sources and
formats.
"""
from .connector import Connector
from .csv_connector import CsvConnector
from .file_connector import FileConnector
from .json_connector import JsonConnector
from .parquet_connector import ParquetConnector
from .base_connector import BaseConnector
from .connector_builder import ConnectorBuilder


__all__ = [
    "BaseConnector",
    "Connector",
    "FileConnector",
    "CsvConnector",
    "JsonConnector",
    "ParquetConnector",
    "ConnectorBuilder"
]



"""CanCreateMixin module."""
from abc import abstractmethod, ABC


class CanDelete(ABC):
    """Delete rows from a table."""

    @abstractmethod
    def delete(self, query: str) -> None:
        """Delete object from storage."""



"""
Config module
"""
from .config import Config, ConfigBuilder
from .file_config import FileConfig
from .aws_credentials import AwsCredentials
from .csv_config import CsvConfig
from .parquet_config import ParquetConfig
from .json_config import JsonConfig


__all__ = [
    "Config",
    "ConfigBuilder",
    "FileConfig",
    "AwsCredentials",
    "CsvConfig",
    "ParquetConfig",
    "JsonConfig",
]



"""Utility function to format multiple types for printing."""
from functools import singledispatch
from typing import _GenericAlias, get_args, get_origin  # type: ignore
from types import GenericAlias


@singledispatch
def pretty(obj):
    """Pretty prints types."""
    raise NotImplementedError


@pretty.register
def _(tpe: None):
    return ""


@pretty.register
def _(type_name: str) -> str:
    return type_name.split(".")[-1]


@pretty.register
def _(alias: _GenericAlias):
    origin = get_origin(alias)
    origin_str = pretty(origin)
    args = get_args(alias)
    args_str = ", ".join(pretty(a) for a in args)
    return f"{origin_str}[{args_str}]"


@pretty.register
def _(tpe: GenericAlias) -> str:  # pragma: py-gt-310
    origin = get_origin(tpe)
    args = get_args(tpe)
    origin_str = pretty(origin)
    args_str = ", ".join(pretty(a) for a in args)
    return f"{origin_str}[{args_str}]"


@pretty.register
def _(tpe: type) -> str:
    if isinstance(tpe, GenericAlias):  # pragma: py-gt-310
        origin = get_origin(tpe)
        args = get_args(tpe)
        origin_str = pretty(origin)
        args_str = ", ".join(pretty(a) for a in args)
        return f"{origin_str}[{args_str}]"
    return tpe.__name__



"""HasBenchmarkMixin module."""
from abc import abstractmethod, ABC
from pysetl.utils import BenchmarkResult


class HasBenchmark(ABC):
    """Indicate if the execution is benchmarked or not."""

    benchmarked: bool = True

    @abstractmethod
    def get_benchmarks(self) -> list[BenchmarkResult]:
        """Return execution benchmarks."""



"""
ParquetConfig module
"""
from typing_extensions import Self
from pysetl.enums import FileStorage
from .file_config import FileConfig, FileConfigModel


class ParquetConfigModel(FileConfigModel):
    """
    ParquetConfig validator
    """


class ParquetConfig(FileConfig):
    """
    ParquetConfig class
    """
    @property
    def config(self: Self) -> ParquetConfigModel:
        _config: dict = self.params | {"storage": FileStorage.PARQUET}

        return ParquetConfigModel(**_config)



"""Deliverable class."""
from __future__ import annotations
from textwrap import dedent
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic, TypeVar, get_args, Union
from typing_extensions import Self
from pysetl.utils import pretty
from pysetl.utils.mixins import IsIdentifiable
from pysetl.workflow.external import External
from pysetl.workflow.delivery_type import DeliveryType


if TYPE_CHECKING:
    from .factory import Factory  # pragma: no cover


T = TypeVar("T")


@dataclass
class Deliverable(Generic[T], IsIdentifiable):
    """
    Deliverable is a container for data.

    A deliverable will transfer data through a Dispatcher.
    @param payload data that will be transferred
    """

    payload: T
    producer: Union[type[Factory], type[External]] = External
    consumers: set[type[Factory]] = field(default_factory=set)
    delivery_id: str = ""

    def __post_init__(self: Self) -> None:
        """Initialize UUID from IsIdentifiable."""
        IsIdentifiable.__init__(self)

    @property
    def payload_type(self: Self) -> DeliveryType:
        """Returns Deliverable payload type."""
        original_class = getattr(self, "__orig_class__")
        contained_type, *_ = get_args(original_class)

        return DeliveryType(contained_type)

    def same_deliverable(self: Self, other: Deliverable) -> bool:
        """Compare two deliverables."""
        _intersection = self.consumers.intersection(other.consumers)
        same_consumer = len(_intersection) != 0 \
            if len(self.consumers) != 0 or other.consumers != 0 \
            else True

        return self.has_same_payload_type(other) and \
            same_consumer and \
            self.producer == other.producer

    def __str__(self) -> str:
        """Pretty print for a deliverable."""
        consumers_string = ", ".join([pretty(x) for x in self.consumers])
        producer_str = pretty(self.producer)
        payload_type_str = pretty(self.payload_type.tp)

        return dedent(f"""
        Deliverable: {payload_type_str}
          From: {producer_str}
          To: {consumers_string}
        """).strip()

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        payload_type_str = pretty(self.payload_type.tp)

        return f"Deliverable[{payload_type_str}]"

    def has_same_payload_type(self: Self, that: Deliverable) -> bool:
        """Compare types bewteen deliverable payloads."""
        return self.payload_type == that.payload_type

    def has_same_payload_type_name(self: Self, name: str) -> bool:
        """Compare types bewteen deliverable payloads."""
        return pretty(self.payload_type.tp) == name

    def is_payload_type(self: Self, __payload_type: type) -> bool:
        """Check payload type of Deliverable."""
        print("self", self.payload_type, type(self.payload_type))
        print("other", __payload_type, type(__payload_type))

        return self.payload_type == __payload_type

    def set_producer(self: Self, producer: type) -> Self:
        """Setter method for producer private property."""
        self.producer = producer

        return self

    def set_consumers(self: Self, consumers: set[type[Factory]]) -> Self:
        """Setter method for consumers private property."""
        self.consumers = consumers

        return self

    def set_delivery_id(self: Self, delivery_id) -> Self:
        """Setter method for delivery_id private property."""
        self.delivery_id = delivery_id

        return self



"""Inspector module."""
from __future__ import annotations
from uuid import UUID
from itertools import groupby
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
from pysetl.utils.mixins import HasLogger
from pysetl.utils.exceptions import PipelineException
from .dag import DAG
from .node import Node
from .flow import Flow
from .delivery import Delivery
from .external import External
from .node_output import NodeOutput


if TYPE_CHECKING:
    from .factory import Factory  # pragma: no cover
    from .pipeline import Pipeline  # pragma: no cover


class Inspector(HasLogger):
    """
    Inspector for a Pipeline.

    Inspector will inspect a given Pipeline and create a Directed Acyclic Graph
    (DAG) with nodes (factories) and flows (data transfer flows)
    """

    def __init__(self: Self, pipeline: Pipeline):
        self.pipeline = pipeline
        self.nodes: set[Node] = self.create_nodes()
        self.flows: set[Flow] = self.create_flows()

    @property
    def graph(self: Self) -> DAG:
        """Return a DAG from a Pipeline."""
        return DAG(self.nodes, self.flows)

    def find_node(self: Self, factory: Factory) -> Optional[Node]:
        """Find a node by Factory."""
        found_nodes = [n for n in self.nodes if n.factory_uuid == factory.uuid]

        if len(found_nodes) == 0:
            return None

        if len(found_nodes) == 1:
            return found_nodes[0]

        raise PipelineException("Multiple factories found")

    def create_nodes(self: Self) -> set[Node]:
        """Create a set of nodes from pipeline stages."""
        return {n for s in self.pipeline.stages for n in s.create_nodes()}

    def find_internal_flows(self: Self, factory: Factory) -> list[Flow]:
        """Find internal flows between stages."""
        parent_node = self.find_node(factory=factory)

        return [
            Flow(parent_node, target_node)
            for target_node
            in self.nodes
            if parent_node is not None and
            target_node.stage > parent_node.stage and
            parent_node.target_node(target_node)
        ]

    def create_internal_flows(self: Self) -> set[Flow]:
        """Create internal flows from a set of stages."""
        return {
            flow
            for stage
            in self.pipeline.stages
            for factory
            in stage.factories
            for flow
            in self.find_internal_flows(factory)
            if not stage.end
        }

    def possible_internal(
            self: Self,
            internal_flows: set[Flow],
            delivery: Delivery,
            factory_uuid: UUID
            ) -> int:
        """Return length potential internal flows for a given delivery."""
        return len([
            f
            for f
            in internal_flows
            if f.payload == delivery.payload_type and
            f.to_node.factory_uuid == factory_uuid and
            f.delivery_id == delivery.delivery_id
        ])

    def deliveries_with_same_id(
            self: Self,
            deliveries: list[Delivery],
            delivery_id: str) -> int:
        """Find deliveries with the same id."""
        return len([d for d in deliveries if d.delivery_id == delivery_id])

    def find_external_flows(
            self: Self,
            node: Node,
            internal_flows: set[Flow]) -> list[Flow]:
        """Find flows from external sources."""
        grouped_deliveries = [
            list(deliveries)
            for _, deliveries
            in groupby(node.input, lambda _: _.payload_type)
        ]

        return [
            Flow(
                Node.external(output=NodeOutput(
                    delivery_type=delivery.payload_type,
                    consumer=set(),
                    delivery_id=delivery.delivery_id,
                    external=True
                )),
                node
            )
            for deliveries
            in grouped_deliveries
            for delivery
            in deliveries
            if (delivery.producer is External) and (
                self.possible_internal(
                    internal_flows,
                    delivery,
                    node.factory_uuid
                ) == (self.deliveries_with_same_id(
                    deliveries,
                    delivery.delivery_id
                ) - 1)
            )
        ]

    def create_external_flows(
            self: Self,
            internal_flows: set[Flow]) -> set[Flow]:
        """Create flows from external sources."""
        return {
            flow
            for node
            in self.nodes
            for flow
            in self.find_external_flows(node, internal_flows)
        }

    def create_flows(self: Self) -> set[Flow]:
        """Create internal and external flows."""
        internal_flows = self.create_internal_flows()
        external_flows = self.create_external_flows(internal_flows)

        return internal_flows.union(external_flows)

    def inspect(self: Self) -> Self:
        """Inspect a Pipeline and create nodes and flow for the DAG."""
        self.nodes = self.create_nodes()
        self.flows = self.create_flows()

        return self

    def __str__(self: Self) -> str:
        return f"========== Pipeline Summary ==========\n\n{self.graph}"

    def __repr__(self: Self) -> str:
        return str(self)



"""Storage module."""
from .connector.connector_builder import ConnectorBuilder
from .repository.sparkrepository_builder import SparkRepositoryBuilder


__all__ = [
    "ConnectorBuilder",
    "SparkRepositoryBuilder",
]



"""ExpectedDelivery Module."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
from pysetl.utils import pretty
from .factory import Factory


@dataclass
class ExpectedDeliverable:
    """
    Expected Deliverable metadata.

    A Deliverable metadata conatainer to trace available deliverables in a
    Pipeline
    """

    deliverable_type: str
    delivery_id: str
    producer: type
    consumer: Optional[type[Factory]]

    def __hash__(self) -> int:
        return hash(
            self.deliverable_type +
            self.delivery_id +
            pretty(self.producer) +
            pretty(self.consumer)
        )



"""CanCreate module."""
from abc import abstractmethod, ABC
from pyspark.sql import DataFrame


class CanCreate(ABC):
    """Abstract Mixin to add create functionality to a connector."""

    @abstractmethod
    def create(
        self,
        dataframe: DataFrame
    ) -> None:
        """Create DataFrame on storage with optional suffix."""



"""
IsConfigurable Module
"""
from abc import ABC, abstractmethod
from typing import Any
from typing_extensions import Self


class IsConfigurable(ABC):
    """
    IsConfigurable mixin
    """
    @abstractmethod
    def get(self: Self, key: str) -> Any:
        """
        Returns the value of the key in the mapping
        """

    @abstractmethod
    def set(self: Self, key: str, value: Any) -> Self:
        """
        Sets a value to a key in the mapping
        """



"""
Enums module
"""
from .db_storage import DbStorage
from .file_storage import FileStorage
from .spark_encodings import SparkEncoding
from .spark_savemodes import SaveMode


__all__ = [
    "DbStorage",
    "FileStorage",
    "SparkEncoding",
    "SaveMode"
]



"""Pipeline module."""
from typing import Any
from typing_extensions import Self
from pysetl.utils import BenchmarkResult, pretty
from pysetl.utils.exceptions import InvalidDeliveryException, PipelineException
from pysetl.utils.mixins import (
    HasRegistry, HasLogger, IsIdentifiable, HasBenchmark,
    HasDiagram
)
from .factory import Factory
from .stage import Stage
from .deliverable import Deliverable
from .dispatcher import Dispatcher
from .inspector import Inspector
from .expected_deliverable import ExpectedDeliverable
from .external import External


PotentialDeliverables = list[tuple[ExpectedDeliverable, ExpectedDeliverable]]


class Pipeline(
    HasRegistry[Stage], IsIdentifiable,
    HasBenchmark, HasDiagram, HasLogger
):
    """Pipeline is a complete data transformation workflow."""

    def __init__(self, benchmark: bool = False) -> None:
        super().__init__()
        IsIdentifiable.__init__(self)
        self.benchmarked = benchmark
        self.inspector = Inspector(self)
        self.dispatcher = Dispatcher()
        self.__stage_counter: int = 0
        self.__benchmark_result: list[BenchmarkResult] = []

    @property
    def stages(self: Self) -> list[Stage]:
        """Return current stages registered in popeline."""
        return list(self.get_registry().values())

    def set_input_from_deliverable(
            self: Self,
            deliverable: Deliverable) -> Self:
        """Register a deliverable Dispatcher."""
        self.dispatcher.add_deliverable(deliverable)

        return self

    def add_stage(self: Self, stage: Stage) -> Self:
        """Register a stage in pipeline."""
        self.log_debug(f"Add stage {self.__stage_counter}")

        self.reset_end_stage()
        stage.stage_id = len(self.get_registry())
        self.register(stage)

        return self

    def reset_end_stage(self: Self) -> None:
        """Reset end of pipeline."""
        last_item = self.last_registered_item

        if last_item:
            last_item.end = False

    def add_stage_from_factory(self: Self, factory: Factory) -> Self:
        """Register a new stage with a single factory."""
        stage = Stage().add_factory(factory)
        self.add_stage(stage)

        return self

    def add_stage_from_type(
            self: Self,
            factory_type: type[Factory],
            *args, **kwargs) -> Self:
        """Instantiate a factory and register a new stage with it."""
        factory = factory_type(*args, **kwargs)
        stage = Stage().add_factory(factory)
        self.add_stage(stage)

        return self

    def describe(self: Self) -> str:
        """Graph representation of the pipeline."""
        self.inspector = Inspector(self)

        return str(self.inspector)

    def to_diagram(self: Self) -> str:
        """Diagram represention of the pipeline."""
        self.inspector = Inspector(self)

        return self.inspector.inspect().graph.to_diagram()

    @property
    def diagram_id(self: Self) -> str:
        """Pipeline has no diagram id."""
        raise NotImplementedError("Pipeline doesn't have a diagram id")

    def get_pipeline_deliverables(self: Self) -> set[ExpectedDeliverable]:
        """Retrive deliverables set directly on pipeline."""
        consumer_deliverables = {
            ExpectedDeliverable(
                deliverable_type=pretty(deliverable.payload_type.tp),
                delivery_id=deliverable.delivery_id,
                producer=pretty(deliverable.producer),
                consumer=pretty(consumer)
            )
            for deliverable
            in self.dispatcher.get_registry().values()
            for consumer
            in deliverable.consumers
        }

        producer_deliverables = {
            ExpectedDeliverable(
                deliverable_type=pretty(deliverable.payload_type.tp),
                delivery_id=deliverable.delivery_id,
                producer=pretty(deliverable.producer),
                consumer=None
            )
            for deliverable
            in self.dispatcher.get_registry().values()
            if len(deliverable.consumers) == 0
        }

        return consumer_deliverables.union(producer_deliverables)

    def get_stages_deliverables(self: Self) -> set[ExpectedDeliverable]:
        """Retrive deliverables available from stages."""
        consumer_deliverables = {
            ExpectedDeliverable(
                deliverable_type=pretty(factory.delivery_type()),
                delivery_id=factory.delivery_id,
                producer=pretty(type(factory)),
                consumer=pretty(consumer)
            )
            for stage
            in self.stages[:-1]
            for factory
            in stage.factories
            for consumer
            in factory.consumers
        }
        producer_deliverables = {
            ExpectedDeliverable(
                deliverable_type=pretty(factory.delivery_type().tp),
                delivery_id=factory.delivery_id,
                producer=pretty(type(factory)),
                consumer=None
            )
            for stage
            in self.stages[:-1]
            for factory
            in stage.factories
        }

        return consumer_deliverables.union(producer_deliverables)

    def get_needed_deliverables(self: Self) -> set[ExpectedDeliverable]:
        """Retrive expected deliverables from each Pipeline node."""
        return {
            ExpectedDeliverable(
                deliverable_type=pretty(delivery.payload_type),
                delivery_id=delivery.delivery_id,
                producer=pretty(delivery.producer),
                consumer=pretty(delivery.consumer)
            )
            for stage
            in self.stages
            for node
            in stage.create_nodes()
            for delivery
            in node.input
        }

    def __compare_deliverables(
            self: Self,
            needed_deliverables: set[ExpectedDeliverable],
            available_deliverables: set[ExpectedDeliverable]
            ) -> PotentialDeliverables:
        """Find potential deliverables for each needed deliverable."""
        return [
            (needed, available)
            for available
            in available_deliverables
            for needed
            in needed_deliverables
            if available.deliverable_type == needed.deliverable_type and
            available.delivery_id == needed.delivery_id and
            (
                True
                if needed.producer == pretty(External)
                else needed.producer == available.producer
            ) and
            (
                available.consumer == needed.consumer or
                not available.consumer
            )
        ]

    def __exists_solution(
            self: Self,
            needed: ExpectedDeliverable,
            potential_deliverables: PotentialDeliverables) -> None:
        """Validate if dependency solution exists."""
        found_by_consumer = list(filter(
            lambda _: _[1].consumer == needed.consumer and
            _[0].deliverable_type == needed.deliverable_type,
            potential_deliverables
        ))
        found_implicit = list(filter(
            lambda _: _[1].consumer is None and
            _[0].deliverable_type == needed.deliverable_type,
            potential_deliverables
        ))

        if not (found_by_consumer or found_implicit):
            delivery_id_str = (
                "''"
                if (needed.delivery_id == "")
                else needed.delivery_id
            )
            raise InvalidDeliveryException(" ".join([
                f"Deliverable of type {needed.deliverable_type}",
                f"with deliveryId {delivery_id_str}",
                f"produced by {pretty(needed.producer)}",
                f"is expected in {pretty(needed.consumer)}."
            ]))

    def run(self: Self) -> Self:
        """Run a pipeline."""
        self.inspector = Inspector(self).inspect()

        needed_deliverables = self.get_needed_deliverables()
        pipeline_deliverables = self.get_pipeline_deliverables()
        stages_output = self.get_stages_deliverables()
        available_deliverables = pipeline_deliverables.union(stages_output)

        potential_deliverables = self.__compare_deliverables(
            needed_deliverables,
            available_deliverables
        )

        _ = [
            self.__exists_solution(needed, potential_deliverables)
            for needed
            in needed_deliverables
        ]

        self.__benchmark_result = [
            benchmark
            for stage
            in self.stages
            for benchmark
            in self.execute_stage(stage)
        ]

        return self

    def execute_stage(self: Self, stage: Stage) -> list[BenchmarkResult]:
        """Dispatch dependencies and run a stage."""
        self.log_info(str(stage))

        if len(self.dispatcher.get_registry()) != 0:
            list(map(self.dispatcher.dispatch, stage.factories))

        stage.benchmarked = self.benchmarked

        stage.run()
        self.dispatcher.collect_stage_deliverables(stage)

        return stage.get_benchmarks()

    def get_last_output(self: Self) -> Any:
        """Return last output from pipeline."""
        if self.last_registered_item:
            return self.last_registered_item.factories[-1].get()

        return None

    def get_output(self: Self, factory_type: type[Factory]) -> Any:
        """Return output from a given factory."""
        factory = [
            factory
            for stage
            in self.stages
            for factory
            in stage.factories
            if isinstance(factory, factory_type)
        ]

        if len(factory) == 0:
            raise PipelineException(
                f"There isn't any class {pretty(factory_type)}"
            )

        return factory[0].get()

    def get_deliverable(
            self: Self,
            deliverable_type: type) -> list[Deliverable]:
        """Find deliverables of a given type."""
        return self.dispatcher.find_deliverable_by_type(deliverable_type)

    def get_benchmarks(self) -> list[BenchmarkResult]:
        """Return collected benchmarks."""
        return self.__benchmark_result



"""
Provides the SparkRepository, a high-level, type-safe data access layer using a
pysetl.connector.
"""
from typing import TypeVar, get_args, _GenericAlias  # type: ignore

from pyspark import StorageLevel
from pyspark.sql import DataFrame
from pyspark.sql.types import StructType
from typedspark import DataSet, Schema
from typing_extensions import Self

from pysetl.utils import pretty
from pysetl.utils.mixins import HasLogger, HasSparkSession, CanPartition, CanDrop
from pysetl.storage.connector.base_connector import BaseConnector
from pysetl.utils.exceptions import InvalidConnectorException
from .base_repository import BaseRepository


T = TypeVar("T", bound=Schema)


class SparkRepository(BaseRepository[T], HasLogger, HasSparkSession):
    """
    SparkRepository class.

    A SparkRepository garantees read-after-write consistency and
    links a output type T to its connector and config
    """

    def __init__(
        self,
        connector: BaseConnector,
        cache: bool = False,
        flush: bool = False,
        storage_level: StorageLevel = StorageLevel.MEMORY_AND_DISK,
    ) -> None:
        """Initialize SparkRepository."""
        self.__connector = connector
        self.__cache: bool = cache
        self.__flush: bool = flush
        self.__storage_level: StorageLevel = storage_level
        self.__cached: DataFrame = self.spark.createDataFrame([], StructType())

    @property
    def type_annotation(self: Self) -> _GenericAlias:
        """Return type annotation."""
        return getattr(self, "__orig_class__")

    def __str__(self: Self) -> str:
        """Customize str method."""
        return pretty(self.type_annotation)  # pylint: disable=E1101

    def __repr__(self: Self) -> str:
        """Customize repr method."""
        return str(self)

    @property
    def cache(self: Self) -> bool:
        """Whether the dataframe is cached."""
        return self.__cache

    @cache.setter
    def cache(self: Self, __value: bool) -> None:
        """Setter for cache property."""
        self.__cache = __value

    def set_cache(self: Self, __value: bool) -> Self:
        """Setter for cache_property."""
        self.__cache = __value
        return self

    @property
    def connector(self: Self) -> BaseConnector:
        """Exposes repository connector."""
        return self.__connector

    def load(self: Self) -> DataSet[T]:
        """Read all data from data storage."""
        __type, *_ = get_args(self.type_annotation)

        if self.__cache and (self.__flush or not self.__cached.is_cached):
            self.__cached = self.connector.read().persist(self.__storage_level)
            data = self.__cached
        elif self.__cache and not self.__flush and self.__cached.is_cached:
            data = self.__cached
        else:
            data = self.connector.read()

        self.__flush = False

        return DataSet[__type](data)  # type: ignore

    def save(self: Self, data: DataSet[T]) -> Self:
        """Save data."""
        self.__flush = True
        self.connector.write(data)

        return self

    def list_partitions(self: Self) -> list[str]:
        """List current all current available partitions in the storage."""
        if not isinstance(self.connector, CanPartition):
            raise InvalidConnectorException(
                "Current connector doesn't support partitions"
            )

        return self.connector.list_partitions()

    def load_partitions(self: Self, partitions: list[str]) -> DataSet[T]:
        """Retrive a subset of partitions from the data."""
        if not isinstance(self.connector, CanPartition):
            raise InvalidConnectorException(
                "Current connector doesn't support partitions"
            )

        dataframe = self.connector.read_partitions(partitions)
        [__tpe, *_] = get_args(self.type_annotation)

        return DataSet[__tpe](dataframe)  # type: ignore

    def drop(self: Self) -> Self:
        """Drop the entire data. Drop table o remove FS directory."""
        if not isinstance(self.connector, CanDrop):
            raise InvalidConnectorException("Current connector doesn't support drop")

        self.connector.drop()

        return self



"""
Defines the abstract base class for all connectors in the pysetl.connector
package.
"""
from abc import ABCMeta, abstractmethod
from pyspark.sql import DataFrame
from typing_extensions import Self


class BaseConnector(metaclass=ABCMeta):
    """
    Abstract Connector.

    A connector should implement read/write operations. These implementations
    should be architecture and object-type specific
    """

    @abstractmethod
    def read(self: Self) -> DataFrame:
        """
        Read data into a DataFrame.

        Read method should implement specific cases for spark connections in
        order to read data
        """

    @abstractmethod
    def write(self: Self, data: DataFrame) -> None:
        """
        Write data into a DataFrame.

        Write method should implement specific cases for spark
        connections in order to write data
        """
